// https://github.com/danburzo/d3-color-difference Version 0.1.3. Copyright 2018 Dan Burzo.
(function (global, factory) {
	typeof exports === 'object' && typeof module !== 'undefined' ? factory(exports, require('d3-color')) :
	typeof define === 'function' && define.amd ? define(['exports', 'd3-color'], factory) :
	(factory((global.d3 = global.d3 || {}),global.d3));
}(this, (function (exports,d3Color) { 'use strict';

	function euclidean(ca1, ca2, ca3, cb1, cb2, cb3) {
		return Math.sqrt(Math.pow(ca1 - cb1, 2) + Math.pow(ca2 - cb2, 2) + Math.pow(ca3 - cb3, 2));
	}

	function differenceEuclideanRGB(std, smp) {
		std = d3Color.rgb(std); smp = d3Color.rgb(smp);
		return euclidean(std.r, std.g, std.b, smp.r, smp.g, smp.b);
	}

	function differenceEuclideanLab(std, smp) {
		std = d3Color.lab(std); smp = d3Color.lab(smp);
		return euclidean(std.l, std.a, std.b, smp.l, smp.a, smp.b);
	}

	function differenceEuclideanHcl(std, smp) {
		std = d3Color.hcl(std); smp = d3Color.hcl(smp);
		return euclidean(std.h, std.c, std.l, smp.h, smp.c, smp.l);
	}

	function differenceEuclideanHsl(std, smp) {
		std = d3Color.hsl(std); smp = d3Color.hsl(smp);
		return euclidean(std.h, std.s, std.l, smp.h, smp.s, smp.l);
	}

	function differenceEuclideanCubehelix(std, smp) {
		std = d3Color.cubehelix(std); smp = d3Color.cubehelix(smp);
		return euclidean(std.h, std.s, std.l, smp.h, smp.s, smp.l);
	}

	function differenceCie94(kL, K1, K2) {

		kL = kL !== undefined ? kL : 1;
		K1 = K1 !== undefined ? K1 : 0.045;
		K2 = K2 !== undefined ? K2 : 0.015;

		return function(std, smp) {
			var LabStd = d3Color.lab(std);
			var LabSmp = d3Color.lab(smp);

			// Extract Lab values, and compute Chroma
			var lStd = LabStd.l;
			var aStd = LabStd.a;
			var bStd = LabStd.b;
			var cStd = Math.sqrt(aStd * aStd + bStd * bStd);

			var lSmp = LabSmp.l;
			var aSmp = LabSmp.a;
			var bSmp = LabSmp.b;
			var cSmp = Math.sqrt(aSmp * aSmp + bSmp * bSmp);

			var dL2 = Math.pow(lStd - lSmp, 2);
			var dC2 = Math.pow(cStd - cSmp, 2);
			var dH2 = Math.pow(aStd - aSmp, 2) + Math.pow(bStd - bSmp, 2) - dC2;

			return Math.sqrt(
				dL2 / Math.pow(kL, 2) + 
				dC2 / Math.pow(1 + K1 * cStd, 2) +
				dH2 / Math.pow(1 + K2 * cStd, 2)
			);
		}
	}

	var differenceCie94Default = differenceCie94();

	/*
		CIEDE2000 color difference, original Matlab implementation by Gaurav Sharma
		Based on "The CIEDE2000 Color-Difference Formula: Implementation Notes, Supplementary Test Data, and Mathematical Observations" 
		by Gaurav Sharma, Wencheng Wu, Edul N. Dalal in Color Research and Application, vol. 30. No. 1, pp. 21-30, February 2005.
		http://www2.ece.rochester.edu/~gsharma/ciede2000/
	 */

	function differenceCiede2000(kL, kC, kH) {

		kL = kL !== undefined ? kL : 1;
		kC = kC !== undefined ? kC : 1;
		kH = kH !== undefined ? kH : 1;

		return function(std, smp) {
			var LabStd = d3Color.lab(std);
			var LabSmp = d3Color.lab(smp);

			var lStd = LabStd.l;
			var aStd = LabStd.a;
			var bStd = LabStd.b;
			var cStd = Math.sqrt(aStd * aStd + bStd * bStd);

			var lSmp = LabSmp.l;
			var aSmp = LabSmp.a;
			var bSmp = LabSmp.b;
			var cSmp = Math.sqrt(aSmp * aSmp + bSmp * bSmp);

			var cAvg = (cStd + cSmp) / 2;
			
			var G = 0.5 * (1 - Math.sqrt(Math.pow(cAvg, 7) / (Math.pow(cAvg, 7) + Math.pow(25, 7))));

			var apStd = aStd * (1 + G);
			var apSmp = aSmp * (1 + G);
			
			var cpStd = Math.sqrt(apStd * apStd + bStd * bStd);
			var cpSmp = Math.sqrt(apSmp * apSmp + bSmp * bSmp);
			
			var hpStd = Math.abs(apStd) + Math.abs(bStd) === 0 ? 0 : Math.atan2(bStd, apStd);
				hpStd += (hpStd < 0) * 2 * Math.PI;

			var hpSmp = Math.abs(apSmp) + Math.abs(bSmp) === 0 ? 0 : Math.atan2(bSmp, apSmp);
				hpSmp += (hpSmp < 0) * 2 * Math.PI;

			var dL = lSmp - lStd;
			var dC = cpSmp - cpStd;
			
			var dhp = cpStd * cpSmp === 0 ? 0 : hpSmp - hpStd;
				dhp -= (dhp > Math.PI) * 2 * Math.PI;
				dhp += (dhp < -Math.PI) * 2 * Math.PI;

			var dH = 2 * Math.sqrt(cpStd * cpSmp) * Math.sin(dhp / 2);

			var Lp = (lStd + lSmp) / 2;
			var Cp = (cpStd + cpSmp) / 2;

			var hp;
			if (cpStd * cpSmp === 0) {
				hp = hpStd + hpSmp;
			} else {
				hp = (hpStd + hpSmp) / 2;
				hp -= (Math.abs(hpStd - hpSmp) > Math.PI) * Math.PI;
				hp += (hp < 0) * 2 * Math.PI;
			}

			var Lpm50 = Math.pow(Lp - 50, 2);
			var T = 1 - 
					0.17 * Math.cos(hp - Math.PI / 6) + 
					0.24 * Math.cos(2 * hp) + 
					0.32 * Math.cos(3 * hp + Math.PI / 30) -
					0.20 * Math.cos(4 * hp - 63 * Math.PI / 180);
			
			var Sl = 1 + (0.015 * Lpm50) / Math.sqrt(20 + Lpm50);
			var Sc = 1 + 0.045 * Cp;
			var Sh = 1 + 0.015 * Cp * T;
			
			var deltaTheta = 30 * Math.PI / 180 * Math.exp(-1 * Math.pow((180 / Math.PI * hp - 275)/25, 2));
			var Rc = 2 * Math.sqrt(
				Math.pow(Cp, 7) / (Math.pow(Cp, 7) + Math.pow(25, 7))
			);

			var Rt = -1 * Math.sin(2 * deltaTheta) * Rc;

			return Math.sqrt(
				Math.pow(dL / (kL * Sl), 2) + 
				Math.pow(dC / (kC * Sc), 2) + 
				Math.pow(dH / (kH * Sh), 2) + 
				Rt * dC / (kC * Sc) * dH / (kH * Sh)
			);
		};
	}

	var differenceCiede2000Default = differenceCiede2000();

	function differenceCmc(l, c) {

		l = l !== undefined ? l : 1;
		c = c !== undefined ? c : 1;

		return function(std, smp) {

			var LabStd = d3Color.lab(std);
			var LabSmp = d3Color.lab(smp);

			var lStd = LabStd.l;
			var aStd = LabStd.a;
			var bStd = LabStd.b;
			var cStd = Math.sqrt(aStd * aStd + bStd * bStd);
			var hStd = Math.atan2(bStd, aStd);
				hStd = hStd + 2 * Math.PI * (hStd < 0);

			var lSmp = LabSmp.l;
			var aSmp = LabSmp.a;
			var bSmp = LabSmp.b;
			var cSmp = Math.sqrt(aSmp * aSmp + bSmp * bSmp);

			var dL2 = Math.pow(lStd - lSmp, 2);
			var dC2 = Math.pow(cStd - cSmp, 2);
			var dH2 = Math.pow(aStd - aSmp, 2) + Math.pow(bStd - bSmp, 2) - dC2;

			var F = Math.sqrt(Math.pow(cStd, 4) / (Math.pow(cStd, 4) + 1900));
			var T = hStd >= (164 / 180 * Math.PI) && 
					hStd <= (345 / 180 * Math.PI) ?
						0.56 + Math.abs(0.2 * Math.cos(hStd + 168 / 180 * Math.PI)) 
						: 0.36 + Math.abs(0.4 * Math.cos(hStd + 35 / 180 * Math.PI));

			var Sl = lStd < 16 ? 0.511 : ((0.040975 * lStd) / (1 + 0.01765 * lStd));
			var Sc = 0.0638 * cStd / (1 + 0.0131 * cStd) + 0.638;
			var Sh = Sc * (F * T + 1 - F);

			return Math.sqrt(
				dL2 / Math.pow(l * Sl, 2),
				dC2 / Math.pow(c * Sc, 2),
				dH2 / Math.pow(Sh, 2)
			);
		};
	}

	var differenceCmcDefault = differenceCmc();

	var θ = 26 / 180 * Math.PI;
	var cosθ = Math.cos(θ);
	var sinθ = Math.sin(θ);
	var factor = 100/Math.log(139/100); // ~ 303.67

	function din99o(color, kCH, kE) {
		var l = factor / kE * Math.log(1 + 0.0039 * color.l);
		if (color.a === 0 && color.b === 0) {
			return d3Color.lab(l, 0, 0); // achromatic colors
		}
		var e = color.a * cosθ + color.b * sinθ;
		var f = 0.83 * (color.b * cosθ - color.a * sinθ);
		var G = Math.sqrt(e * e + f * f);
		var c = Math.log(1 + 0.075 * G) / (0.0435 * kCH * kE);
		var h = (Math.atan2(f, e) + θ) / Math.PI * 180;
		return d3Color.lab(l, c * Math.cos(h / 180 * Math.PI), c * Math.sin(h / 180 * Math.PI));
	}

	function differenceDin99o(kCH, kE) {

		kCH = kCH !== undefined ? kCH : 1;
		kE = kE !== undefined ? kE : 1;

		return function(std, smp) {
			std = din99o(d3Color.lab(std), kCH, kE);
			smp = din99o(d3Color.lab(smp), kCH, kE);
			return euclidean(std.l, std.a, std.b, smp.l, smp.a, smp.b);
		}
	}

	var differenceDin99oDefault = differenceDin99o();

	exports.differenceEuclideanRGB = differenceEuclideanRGB;
	exports.differenceEuclideanLab = differenceEuclideanLab;
	exports.differenceEuclideanHcl = differenceEuclideanHcl;
	exports.differenceEuclideanHsl = differenceEuclideanHsl;
	exports.differenceEuclideanCubehelix = differenceEuclideanCubehelix;
	exports.differenceCie76 = differenceEuclideanLab;
	exports.differenceCie94 = differenceCie94Default;
	exports.differenceCie94Weighted = differenceCie94;
	exports.differenceCiede2000 = differenceCiede2000Default;
	exports.differenceCiede2000Weighted = differenceCiede2000;
	exports.differenceCmc = differenceCmcDefault;
	exports.differenceCmcWeighted = differenceCmc;
	exports.differenceDin99o = differenceDin99oDefault;
	exports.differenceDin99oWeighted = differenceDin99o;

	Object.defineProperty(exports, '__esModule', { value: true });

})));
